import numpy as np
from mpmath import besselj, mpf, pi, sqrt

from ..funcs import j1_over_x


@np.vectorize
def mpmath_jn_over_x(i, x):
    xx = mpf(x)
    if x == 0:
        return float(1 / mpf(3))
    else:
        return float(sqrt(pi / mpf(2) / xx) * besselj(i + mpf("1/2"), xx) / xx)


def test_j1_over_x():
    x = np.concatenate(([0], np.geomspace(1e-8, 10, 1000)))

    yref = mpmath_jn_over_x(1, x)
    yval = j1_over_x(x)

    np.testing.assert_allclose(yref, yval, rtol=1e-14)



"""Projections."""
import numpy as np
from numba import njit

from .decorators import spatial
from .types import FloatArrayType


@spatial
@njit
def cic_projection(pos: FloatArrayType, N: int) -> FloatArrayType:
    """Compute a CIC projection of some position on a uniform grid.

    Parameters
    ----------
    pos : array-like (Npoint, 3)
        The position in the [0,1] periodic domain
    N : int
        The size of the grid to project onto

    Returns
    -------
    dens : array-like (N, N, N)
        The density.

    Notes
    -----
    The scheme conserves the volume, so that dens.sum() == Npoint.
    The density array follows the C convention, i.e. it is ordered as z, y, x

    Example
    -------
        X = np.array([
            [0.42, 0.125, 0.125],
            [0.51, 0.125, 0.125]])
        dens = cic_projection(X, 4)

        # Varying dimension
        #    z
        #    |  y
        #    |  |  x
        dens[0, 0, :] == [0, 1.28, 0.72, 0]
    """
    posN = pos * N
    dens = np.zeros((N, N, N))
    for ipart in range(len(pos)):
        x, y, z = posN[ipart, 0], posN[ipart, 1], posN[ipart, 2]
        i = int(np.floor(x - 0.5))
        j = int(np.floor(y - 0.5))
        k = int(np.floor(z - 0.5))

        vi0 = min(i + 1, x + 0.5) - max(i, x - 0.5)
        vj0 = min(j + 1, y + 0.5) - max(j, y - 0.5)
        vk0 = min(k + 1, z + 0.5) - max(k, z - 0.5)

        vi1 = 1 - vi0
        vj1 = 1 - vj0
        vk1 = 1 - vk0

        for kk, vk in zip((k, k + 1), (vk0, vk1)):
            for jj, vj in zip((j, j + 1), (vj0, vj1)):
                for ii, vi in zip((i, i + 1), (vi0, vi1)):
                    v = vi * vj * vk
                    dens[kk % N, jj % N, ii % N] += v

    return dens


@spatial
@njit
def nearest_projection(pos: FloatArrayType, N: int) -> FloatArrayType:
    """Compute a nearest projection of some position on a uniform grid.

    Parameters
    ----------
    pos : array-like (Npoint, 3)
        The position in the [0,1] periodic domain
    N : int
        The size of the grid to project onto

    Returns
    -------
    dens : array-like (N, N, N)
        The density.

    Notes
    -----
    The scheme conserves the volume, so that dens.sum() == Npoint.
    """
    posN = pos * N
    dens = np.zeros((N, N, N))
    for ipart in range(len(pos)):
        x, y, z = posN[ipart, 0], posN[ipart, 1], posN[ipart, 2]
        i = int(np.floor(x - 0.5))
        j = int(np.floor(y - 0.5))
        k = int(np.floor(z - 0.5))

        dens[k % N, j % N, i % N] += 1

    return dens



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from kaitaistruct import KaitaiStruct, __version__ as ks_version
from pkg_resources import parse_version

if parse_version(ks_version) < parse_version("0.7"):
    raise Exception(
        "Incompatible Kaitai Struct Python API: 0.7 or later is required, but you have %s"
        % (ks_version)
    )


class DisperseReader(KaitaiStruct):
    """
    .. seealso::
       Source - http://www2.iap.fr/users/sousbie/web/html/indexf83e.html?post/NDskl-format
    """

    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.header = self._root.Header(self._io, self, self._root)
        self.data = self._root.Data(self._io, self, self._root)

    class SegDataStruct(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.next_node = self._io.read_u4le()
            self.next_seg = self._io.read_u4le()

    class Data(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.seg_pos = [None] * (
                (2 * self._root.header.ndim) * self._root.header.nseg
            )
            for i in range((2 * self._root.header.ndim) * self._root.header.nseg):
                self.seg_pos[i] = self._io.read_f4le()

            self._unnamed1 = self._io.read_u4le()
            self._unnamed2 = self._io.read_u4le()
            self.node_pos = [None] * (self._root.header.ndim * self._root.header.nnode)
            for i in range(self._root.header.ndim * self._root.header.nnode):
                self.node_pos[i] = self._io.read_f4le()

            self._unnamed4 = self._io.read_u4le()
            self._unnamed5 = self._io.read_u4le()
            self.seg_data = [None] * (
                self._root.header.nseg * self._root.header.nseg_data
            )
            for i in range(self._root.header.nseg * self._root.header.nseg_data):
                self.seg_data[i] = self._io.read_f8le()

            self._unnamed7 = self._io.read_u4le()
            self._unnamed8 = self._io.read_u4le()
            self.node_data = [None] * (
                self._root.header.nnode * self._root.header.nnode_data
            )
            for i in range(self._root.header.nnode * self._root.header.nnode_data):
                self.node_data[i] = self._io.read_f8le()

            self._unnamed10 = self._io.read_u4le()
            self._unnamed11 = self._io.read_u4le()
            self.node_data_int = [None] * (self._root.header.nnode)
            for i in range(self._root.header.nnode):
                self.node_data_int[i] = self._root.NodeStruct(
                    self._io, self, self._root
                )

            self._unnamed13 = self._io.read_u4le()
            self._unnamed14 = self._io.read_u4le()
            self.seg_data_int = [None] * (self._root.header.nseg)
            for i in range(self._root.header.nseg):
                self.seg_data_int[i] = self._root.SegStruct(self._io, self, self._root)

            self._unnamed16 = self._io.read_u4le()

    class Header(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self._unnamed0 = self._io.ensure_fixed_contents(b"\x10\x00\x00\x00")
            self._unnamed1 = self._io.ensure_fixed_contents(
                b"\x4E\x44\x53\x4B\x45\x4C\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            )
            self._unnamed2 = self._io.ensure_fixed_contents(
                b"\x10\x00\x00\x00\x00\x00\x00\x00"
            )
            self.comment = (self._io.read_bytes(80)).decode("ASCII")
            self.ndim = self._io.read_u4le()
            self.dims = [None] * (self.ndim)
            for i in range(self.ndim):
                self.dims[i] = self._io.read_u4le()

            self._unnamed6 = [None] * (20 - self.ndim)
            for i in range(20 - self.ndim):
                self._unnamed6[i] = self._io.read_u4le()

            self.x0 = [None] * (self.ndim)
            for i in range(self.ndim):
                self.x0[i] = self._io.read_f8le()

            self._unnamed8 = [None] * (20 - self.ndim)
            for i in range(20 - self.ndim):
                self._unnamed8[i] = self._io.read_f8le()

            self.delta = [None] * (self.ndim)
            for i in range(self.ndim):
                self.delta[i] = self._io.read_f8le()

            self._unnamed10 = [None] * (20 - self.ndim)
            for i in range(20 - self.ndim):
                self._unnamed10[i] = self._io.read_f8le()

            self.nseg = self._io.read_u4le()
            self.nnode = self._io.read_u4le()
            self.nseg_data = self._io.read_u4le()
            self.nnode_data = self._io.read_u4le()
            self._unnamed15 = self._io.read_u4le()
            self._unnamed16 = self._io.read_u4le()
            self.seg_data_info = [None] * (self.nseg_data)
            for i in range(self.nseg_data):
                self.seg_data_info[i] = (self._io.read_bytes(20)).decode("ASCII")

            self._unnamed18 = self._io.read_u4le()
            self._unnamed19 = self._io.read_u4le()
            self.node_data_info = [None] * (self.nnode_data)
            for i in range(self.nnode_data):
                self.node_data_info[i] = (self._io.read_bytes(20)).decode("ASCII")

            self._unnamed21 = self._io.read_u4le()
            self._unnamed22 = self._io.read_u4le()

    class NodeStruct(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.pos_index = self._io.read_u4le()
            self.flags = self._io.read_u4le()
            self.nnext = self._io.read_u4le()
            self.type = self._io.read_u4le()
            self.index = self._io.read_u4le()
            self.nseg = [None] * (self.nnext)
            for i in range(self.nnext):
                self.nseg[i] = self._io.read_u4le()

            self.seg_data = [None] * (self.nnext)
            for i in range(self.nnext):
                self.seg_data[i] = self._root.SegDataStruct(self._io, self, self._root)

    class SegStruct(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.pos_index = self._io.read_u4le()
            self.node_ids = [None] * (2)
            for i in range(2):
                self.node_ids[i] = self._io.read_u4le()

            self.flags = self._io.read_u4le()
            self.index = self._io.read_u4le()
            self.next_seg = self._io.read_u4le()
            self.prev_seg = self._io.read_u4le()



"""Types used throughout the project"""

import pathlib
from typing import Any, Union

PathType = Union[str, pathlib.Path]
IntArrayType = Any  # np.ndarray[int]
FloatArrayType = Any  # np.ndarray[float]



"""Scale bar from https://gist.github.com/dmeliza/3251476."""
# Adapted from mpl_toolkits.axes_grid1
# LICENSE: Python Software Foundation (http://docs.python.org/license.html)

import warnings

import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.mathtext import MathTextWarning
from matplotlib.offsetbox import AnchoredOffsetbox


class AnchoredScaleBar(AnchoredOffsetbox):
    def __init__(
        self,
        transform,
        sizex=0,
        sizey=0,
        labelx=None,
        labely=None,
        loc=4,
        pad=0.1,
        borderpad=0.1,
        sep=2,
        prop=None,
        barcolor="black",
        barwidth=None,
        label_kwa=None,
        **kwargs,
    ):
        """
        Draw a horizontal and/or vertical  bar with the size in data coordinate
        of the give axes. A label will be drawn underneath (center-aligned).

        Parameters
        ----------
        - transform : the coordinate frame (typically axes.transData)
        - sizex,sizey : width of x,y bar, in data units. 0 to omit
        - labelx,labely : labels for x,y bars; None to omit
        - loc : position in containing axes
        - pad, borderpad : padding, in fraction of the legend font size
          (or prop)
        - sep : separation between labels and bars in points.
        - **kwargs : additional arguments passed to base class constructor
        """
        from matplotlib.offsetbox import AuxTransformBox, HPacker, TextArea, VPacker
        from matplotlib.patches import Rectangle

        if label_kwa is None:
            label_kwa = {}

        bars = AuxTransformBox(transform)
        rect_kwa = {"ec": barcolor, "lw": barwidth, "fc": "none"}
        if sizex:
            bars.add_artist(Rectangle((0, 0), sizex, 0, **rect_kwa))
        if sizey:
            bars.add_artist(Rectangle((0, 0), 0, sizey, **rect_kwa))

        vpacker_kwa = {"align": "center", "pad": 0, "sep": sep}
        if sizex and labelx:
            self.xlabel = TextArea(labelx, minimumdescent=False, **label_kwa)
            bars = VPacker(children=[bars, self.xlabel], **vpacker_kwa)
        if sizey and labely:
            self.ylabel = TextArea(labely, **label_kwa)
            bars = HPacker(children=[self.ylabel, bars], **vpacker_kwa)

        AnchoredOffsetbox.__init__(
            self,
            loc,
            pad=pad,
            borderpad=borderpad,
            child=bars,
            prop=prop,
            frameon=False,
            **kwargs,
        )


def add_scalebar(ax, matchx=True, matchy=True, hidex=True, hidey=True, **kwargs):
    """Add scalebars to axes
    Adds a set of scale bars to *ax*, matching the size to the ticks of the
    plot and optionally hiding the x and y axes.

    Parameters
    ----------
    - ax : the axis to attach ticks to
    - matchx,matchy : if True, set size of scale bars to spacing between ticks
                    if False, size should be set using sizex and sizey params
    - hidex,hidey : if True, hide x-axis and y-axis of parent
    - **kwargs : additional arguments passed to AnchoredScaleBars

    Returns:
    --------
    sb : scalebar object
    """

    def f(axis):
        locs = axis.get_majorticklocs()
        return len(locs) > 1 and (locs[1] - locs[0])

    if matchx:
        kwargs["sizex"] = f(ax.xaxis)
        kwargs["labelx"] = str(kwargs["sizex"])
    if matchy:
        kwargs["sizey"] = f(ax.yaxis)
        kwargs["labely"] = str(kwargs["sizey"])

    sb = AnchoredScaleBar(ax.transData, **kwargs)
    ax.add_artist(sb)

    if hidex:
        ax.xaxis.set_visible(False)
    if hidey:
        ax.yaxis.set_visible(False)
    if hidex and hidey:
        ax.set_frame_on(False)

    return sb


def fix_glyph_errors(ax=None):
    """Fix missing glyph warnings in matplotlib"""
    # From https://stackoverflow.com/a/47850541

    if mpl.rcParams["axes.unicode_minus"] is False:
        warnings.warn(
            "If you have issues with minus sign, set `axes.unicode_minus` to True in the rcParams."
        )
    if mpl.rcParams["text.usetex"]:
        # Everything is handled by LaTeX, so nothing to do.
        return

    if ax is None:
        ax = plt.gca()
    fig = ax.get_figure()
    # Force the figure to be drawn
    if tuple(int(_) for _ in mpl.__version__.split(".")) >= (3, 1, 0):
        import logging

        logger = logging.getLogger("matplotlib.mathtext")
        original_level = logger.getEffectiveLevel()
        logger.setLevel(logging.ERROR)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=MathTextWarning)
            fig.canvas.draw()
        logger.setLevel(original_level)
    else:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=MathTextWarning)
            fig.canvas.draw()
    # Remove '\mathdefault' from all minor tick labels
    labels = [
        label.get_text().replace(r"\mathdefault", "")
        for label in ax.get_xminorticklabels()
    ]
    ax.set_xticklabels(labels, minor=True)
    labels = [
        label.get_text().replace(r"\mathdefault", "")
        for label in ax.get_yminorticklabels()
    ]
    ax.set_yticklabels(labels, minor=True)


def adjust_lightness(color, amount=0.5):
    """Adjust the lightness of a colour.

    Parameters
    ----------
    color : matplotlib color ("black", "C0", etc.) or RGB(A) value
    amount : float
        Amplitude of the adjustment. If <1, will darken and if >1, will lighten.
    """
    import colorsys

    import matplotlib.colors as mc

    try:
        c = mc.cnames[color]
    except KeyError:
        c = color
    c = colorsys.rgb_to_hls(*mc.to_rgb(c))
    return colorsys.hls_to_rgb(c[0], max(0, min(1, amount * c[1])), c[2])


def colorbar(mappable, ax=None):
    """Add a colorbar attached to the mappable, using the `make_axes_locatable` API

    Adapted from https://joseph-long.com/writing/colorbars/
    """
    import matplotlib.pyplot as plt
    from mpl_toolkits.axes_grid1 import make_axes_locatable

    try:
        # Save previous current axis
        last_axes = plt.gca()
        ax = mappable.axes if ax is None else ax
        fig = ax.figure
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cbar = fig.colorbar(mappable, cax=cax)
    finally:
        # Restore previous current axis
        plt.sca(last_axes)
    return cbar



"""Useful decorators."""

import os
from functools import wraps
from typing import Callable

import numpy as np

from .exceptions import AstroToolsetNotSpatialError


class read_files:
    """Decorator for functions that take a file as input.

    Parameters:
    -----------
    N : int
        List of the name of the argument that are files

    Notes
    -----
    The wrapped function is expected to have its input files as last argument.
    """

    N_files = 0

    def __init__(self, N: int):
        self.N_files = N

    def __call__(self, fun: Callable) -> Callable:
        @wraps(fun)
        def wrapped(*args, **kwargs):
            nargs = len(args)
            for i in range(self.N_files):
                if not os.path.exists(args[nargs - i - 1]):
                    raise FileNotFoundError(args[nargs - i - 1])

            return fun(*args, **kwargs)

        return wrapped


def spatial(fun):
    """Asserts that the first argument passed to the function is of spatial type,
    i.e. its last dimension has three components.
    """

    @wraps(fun)
    def wrapped(spatial_array, *args, **kwargs):
        try:
            assert isinstance(spatial_array, np.ndarray)
            assert spatial_array.ndim >= 2
            assert spatial_array.shape[-1] == 3
        except AssertionError:
            raise AstroToolsetNotSpatialError(spatial_array)
        return fun(spatial_array, *args, **kwargs)

    return wrapped



#!/home/ccc/anaconda3/bin/python
# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.3.0
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# Writing myself the files

import argparse
import os
import shutil
from collections import defaultdict

# %%
from glob import glob

import ipyplot
import matplotlib as mpl
import numpy as np
import yt
from cython_fortran_file import FortranFile as FF
from scipy.interpolate import interp1d
from tqdm.auto import tqdm
from yt.frontends.ramses.definitions import ramses_header
from yt.frontends.ramses.field_handlers import (
    FieldFileHandler,
    GravFieldFileHandler,
    HydroFieldFileHandler,
)
from yt.frontends.ramses.hilbert import hilbert3d as hilbert3d_yt
from yt.frontends.ramses.particle_handlers import (
    DefaultParticleFileHandler,
    ParticleFileHandler,
    SinkParticleFileHandler,
)

from astrophysics_toolset.ramses.hilbert import hilbert3d
from astrophysics_toolset.ramses.oct_handler import Octree

# %%
parser = argparse.ArgumentParser()
# parser.add_argument('--input', type=str, default='/home/ccc/Documents/Postdoc/genetIC-angular-momentum-constrain/simulations/data/DM_256_planck/resim/00140/halo_189/relative_change_lx_0.8/output_00001/info_00001.txt',
#     help='Input file'
# )
parser.add_argument(
    "--input", required=True, type=str, help="Path to intut info_xxxx.txt file"
)
parser.add_argument(
    "--output-dir", required=True, type=str, default="Path where to write the output"
)
parser.add_argument("--ncpu", type=int, default=4, help="The desired number of CPUs")
parser.add_argument(
    "--longint",
    action="store_true",
    help="Set if you compiled RAMSES with the LONGINT option.",
)
parser.add_argument(
    "--quadhilbert",
    action="store_true",
    help="Set if you compiled RAMSES with the QUADHILBERT option.",
)
parser.add_argument(
    "--nexpand",
    default=1,
    type=int,
    help="Number of times to expand boundaries (default: %(default)s)",
)
parser.add_argument("--test", action="store_true")
parser.add_argument(
    "--ngridmax",
    type=int,
    default=0,
    help="The new value of ngridmax. If set to 0 (default), then compute it from the total number of octs. If set to -1, use the old value and if set to anything other, use this new value",
)
parser.add_argument(
    "--remap",
    choices=("at_boundaries", "interp", "linear"),
    default="at_boundaries",
    help="""Choice of Strategy to compute the new refinement map (default: %(default)s).
        *at_boundaries*: the new domains will be at the old domain boundaries.
        *interp*: linearly interpolate the new hilbert keys from the old map.
        *linear*: regularly space the keys.
    """,
)


def rawQuadToDouble(raw: bytes):
    # From https://stackoverflow.com/a/52568477/2601223
    asint = int.from_bytes(raw, byteorder="little")
    sign = (-1.0) ** (asint >> 127)
    exponent = ((asint >> 112) & 0x7FFF) - 16383
    significand = (asint & ((1 << 112) - 1)) | (1 << 112)
    return sign * significand * 2.0 ** (exponent - 112)


def read_amr(amr_file: str, longint: bool, quadhilbert: bool):
    i8b = "l" if longint else "i"
    qdp = "float128" if quadhilbert else "float64"
    dp = "float64"
    with FF(amr_file) as f:
        headers = {}
        for h in ramses_header(headers):
            headers.update(f.read_attrs(h))

        # Read level variables
        shape = headers["nlevelmax"], headers["ncpu"]
        headl = f.read_vector("i").reshape(shape)
        taill = f.read_vector("i").reshape(shape)
        numbl = f.read_vector("i").reshape(shape)
        try:
            numbtot = f.read_vector(i8b).reshape((10, headers["nlevelmax"]))
        except ValueError:
            raise Exception(
                "Caught an exception while reading numbtot. This is likely due "
                "to you forgetting to (un)set the longint flag!\n"
                "Try calling the script with/out `--longint`."
            )

        # Free memory
        headf, tailf, numbf, used_mem, used_mem_tot = f.read_vector("i")

        # Ordering
        ordering = "".join(f.read_vector("c").astype(str)).strip()
        ndomain = headers["ncpu"] * 1
        if ordering != "hilbert":
            raise NotImplementedError
        try:
            if quadhilbert:
                with open(amr_file, "br") as f2:
                    f2.seek(f.tell())
                    s1 = int.from_bytes(f2.read(4), byteorder="little")
                    bound_keys = np.array(
                        [rawQuadToDouble(f2.read(16)) for i in range(ndomain + 1)],
                        dtype=np.float128,
                    )
                    s2 = int.from_bytes(f2.read(4), byteorder="little")
                    assert s1 == s2
                    f.seek(f2.tell())
            else:
                bound_keys = f.read_vector(qdp).reshape(ndomain + 1)
        except ValueError:
            raise Exception(
                "Caught an exception while reading hilbert keys. This is likely due "
                "to you forgetting to (un)set the quadhibert flag!\n"
                "Try calling the script with/out `--quadhilbert`."
            )

        nlevelmax, nboundary, ncpu, ndim, ngridmax = (
            headers[k] for k in ("nlevelmax", "nboundary", "ncpu", "ndim", "ngridmax")
        )

        # Allocate memory
        _level, _ndom, ind_grid, next, prev, parent = (
            np.zeros(ngridmax, dtype="i") for _ in range(6)
        )
        _level_cell = np.zeros((ngridmax, 8), dtype="i")
        xc = np.zeros((ngridmax, ndim), dtype="d")
        nbor = np.zeros((ngridmax, 2 * ndim), dtype="i")
        son, cpu_map, refmap = (
            np.zeros((ngridmax, 2 ** ndim), dtype="i") for _ in range(3)
        )

        # Coarse levels
        # should be of length 1 unless there are non-periodic boundaries
        ncoarse = 1
        coarse_son = f.read_vector("i").reshape(ncoarse)
        coarse_refmap = f.read_vector("i").reshape(ncoarse)
        coarse_cpu_map = f.read_vector("i").reshape(ncoarse)

        ret = {"headers": headers}
        ret["headl"] = headl
        ret["taill"] = taill
        ret["numbl"] = numbl
        ret["numbtot"] = numbtot
        ret["bound_keys"] = bound_keys
        ret["son"] = son
        ret["refmap"] = refmap
        ret["cpu_map"] = cpu_map

        # Fine levels
        i = 0
        for ilevel in range(nlevelmax):
            for ibound in range(nboundary + ncpu):
                if ibound <= ncpu:
                    ncache = numbl[ilevel, ibound]  # NOTE: C-order vs. F-order
                else:
                    raise NotImplementedError(
                        "Cannot load datasets with non-periodic boundary conditions"
                    )
                    # ncache = numbb[ilevel, ibound - ncpu]

                def read(kind):
                    tmp = f.read_vector(kind)
                    return tmp.reshape(ncache)

                if ncache > 0:
                    ind_grid[i : i + ncache] = read("i")
                    next[i : i + ncache] = read("i")
                    prev[i : i + ncache] = read("i")
                    xc[i : i + ncache] = np.stack(
                        [read(dp) for i in range(ndim)], axis=-1
                    )
                    parent[i : i + ncache] = read("i")
                    nbor[i : i + ncache] = np.stack(
                        [read("i") for idim in range(2 * ndim)], axis=-1
                    )
                    son[i : i + ncache] = np.stack(
                        [read("i") for idim in range(2 ** ndim)], axis=-1
                    )
                    cpu_map[i : i + ncache] = np.stack(
                        [read("i") for idim in range(2 ** ndim)], axis=-1
                    )
                    refmap[i : i + ncache] = np.stack(
                        [read("i") for idim in range(2 ** ndim)], axis=-1
                    )

                    _level[i : i + ncache] = ilevel + 1  # Fortran is 1-indexed
                    _level_cell[i : i + ncache, :] = ilevel + 2
                    _ndom[i : i + ncache] = ibound + 1  # Fortran is 1-indexed
                    i += ncache

        (
            ind_grid,
            next,
            prev,
            xc,
            parent,
            nbor,
            son,
            cpu_map,
            refmap,
            _level,
            _level_cell,
            _ndom,
        ) = (
            _[:i]
            for _ in (
                ind_grid,
                next,
                prev,
                xc,
                parent,
                nbor,
                son,
                cpu_map,
                refmap,
                _level,
                _level_cell,
                _ndom,
            )
        )

        ret.update(
            dict(
                ind_grid=ind_grid,
                next=next,
                prev=prev,
                xc=xc,
                parent=parent,
                nbor=nbor,
                son=son,
                cpu_map=cpu_map,
                refmap=refmap,
                coarse_refmap=coarse_refmap,
                coarse_cpu_map=coarse_cpu_map,
                coarse_son=coarse_son,
                _level=_level,
                _level_cell=_level_cell,
                _ndom=_ndom,
            )
        )

        return ret


def read_all_amr_files(dirname: str, longint: bool, quadhilbert: bool):
    pattern = os.path.join(dirname, "amr_?????.out?????")
    amr_files = sorted(glob(pattern))

    data = {}
    for amr_file in tqdm(amr_files, desc="Reading AMR"):
        icpu = int(amr_file.split(".out")[1])
        data[icpu] = read_amr(amr_file, longint, quadhilbert)  # CPU are indexed by one
    return data


def main():
    args = parser.parse_args()

    input_dir = os.path.abspath(os.path.split(args.input)[0])
    output_dir = os.path.abspath(args.output_dir)
    if input_dir == output_dir or os.path.exists(args.output_dir):
        raise Exception(
            (
                "Either the destination folder already exists (%s) or "
                "you are trying to write new files in the same folder as the old one (%s).\n"
                "Either way, I am Refusing to do that.\n\n\n"
                "Sorry."
            )
            % (output_dir, input_dir)
        )

    # Making output dir
    _iout = int(os.path.split(args.input)[1].split(".txt")[0].split("_")[1])
    CONFIG = dict(new_ncpu=args.ncpu, iout=_iout)

    # %%
    ipos = np.random.randint(0, 2 ** 9, size=(2000, 3))

    a = hilbert3d(ipos, 10)
    b = hilbert3d_yt(ipos, 10)
    np.testing.assert_allclose(a, b)

    ds = yt.load(args.input)
    default_headers = {"ramses": {}, "gravity": {"nvar": 4}}

    path = os.path.split(ds.parameter_filename)[0]
    data = read_all_amr_files(path, args.longint, args.quadhilbert)

    nlevelmin = ds.parameters["levelmin"]
    nlevelmax = data[1]["headers"]["nlevelmax"]
    boxlen = data[1]["headers"]["boxlen"]
    nx_loc = 1
    scale = boxlen / nx_loc
    ndim = data[1]["headers"]["ndim"]
    assert ndim == 3

    bscale = 2 ** (nlevelmax + 1)
    ncode = nx_loc * int(bscale)
    bscale = int(bscale / scale)

    bit_length = -1
    for _bit_length in range(1, 32 + 1):
        ncode = ncode // 2
        if ncode <= 1:
            break
    bit_length = _bit_length
    # At this stage, we have loaded the AMR structure and we make sure we are able to recompute the CPU map. This will be useful in the future.
    dd = np.array(
        [(i, j, k) for k in (-1, 1) for j in (-1, 1) for i in (-1, 1)]
    )  # note the ordering here (for Fortran/C ordering)
    dt = data[1]
    ngridmax = dt["headers"]["ngridmax"]
    ncoarse = np.product(dt["headers"]["nx"])
    CONFIG["ncoarse"] = ncoarse
    ii = np.array([i * ngridmax + ncoarse for i in range(8)])

    Ncell_tot = 0
    for icpu, dt in data.items():
        dx = 1 / 2 / 2 ** dt["_level"]
        ixcell = (
            (dt["xc"][:, None, :] + dx[:, None, None] * dd[None, :, :]) * bscale
        ).astype(int)
        dt["_ixgrid"] = (dt["xc"][:, None, :] * bscale).astype(int)
        dt["_ixcell"] = ixcell
        dt["_hilbert_key"] = hilbert3d(ixcell.reshape(-1, 3), bit_length)
        dt["_hilbert_key_grid"] = hilbert3d(dt["_ixgrid"].reshape(-1, 3), bit_length)

        cpu_map = dt["cpu_map"].reshape(-1)
        cpu_map_with_keys = np.searchsorted(
            dt["bound_keys"], dt["_hilbert_key"], side="left"
        )

        # Since we're casting the keys to float64, some cells will be
        # in the wrong domain with quadhilbert. Make sure that's a minority.
        if args.quadhilbert:
            count = (cpu_map != cpu_map_with_keys).sum()
            if count > dt["headers"]["ncpu"]:
                print(
                    "At this point I am computing the new CPU map and making sure the "
                    "calculations agree with what RAMSES dumped. Unfortunately, I found"
                    f"{count} cells which were attributed to the wrong CPU.\t"
                    "Just wanted to let you know!."
                )
        else:
            assert np.all(cpu_map == cpu_map_with_keys)

        dt["_ind_cell"] = dt["ind_grid"][:, None] + ii[None, :]

        Ncell_tot += (dt["cpu_map"] == icpu).sum()

    # Recompute CPU map using new keys
    old_ncpu = dt["headers"]["ncpu"]
    if args.remap == "interp":
        bound_key_orig = interp1d(
            np.arange(dt["headers"]["ncpu"] + 1), dt["bound_keys"]
        )

        new_bound_keys = bound_key_orig(
            np.linspace(0, dt["headers"]["ncpu"], CONFIG["new_ncpu"] + 1)
        )
    elif args.remap == "linear":
        new_bound_keys = np.linspace(0, dt["bound_keys"][-1], CONFIG["new_ncpu"] + 1)
    elif args.remap == "at_boundaries":
        new_ncpu = CONFIG["new_ncpu"]
        old_bound_keys = dt["bound_keys"]

        icpu_ind = np.floor(np.linspace(0, 1, new_ncpu + 1) * old_ncpu).astype(int)
        old_cpu_to_new_cpu = np.array(
            [
                np.argwhere(icpu_ind == icpu)[0, 0] + 1 if icpu in icpu_ind else 0
                for icpu in range(old_ncpu)
            ]
        )
        new_bound_keys = old_bound_keys[icpu_ind]
        old_cpu_to_new_cpu = np.zeros(old_ncpu, dtype=np.int64)

        for i in range(old_ncpu):
            # Find first key in the new keys that's largest than current
            up_key = dt["bound_keys"][i]
            old_cpu_to_new_cpu[i] = np.argwhere(new_bound_keys > up_key).flatten()[0]
    else:
        raise NotImplementedError(
            f"Remap method {args.remap} has not been implemented."
        )

    # Round the new keys to 64-bit floating precision
    original_dtype = new_bound_keys.dtype
    new_bound_keys = (new_bound_keys.astype(np.float64)).astype(original_dtype)

    # %%
    oct = Octree(nlevelmin, nlevelmax, old_ncpu=old_ncpu, new_ncpu=CONFIG["new_ncpu"])
    N2 = 1
    for icpu, dt in tqdm(data.items(), desc="Adding grids to tree"):
        mask = dt["_level"] <= 99999
        ipos = dt["_ixgrid"][mask].reshape(-1, 3)
        file_ind = dt["ind_grid"][mask].astype(np.int64)
        domain_ind = np.searchsorted(
            dt["bound_keys"], dt["_hilbert_key_grid"][mask], side="left"
        )
        if args.remap == "at_boundaries":
            new_domain_ind = old_cpu_to_new_cpu[domain_ind - 1]
            new_cpu_map = old_cpu_to_new_cpu[dt["cpu_map"] - 1]
        else:
            new_domain_ind = np.searchsorted(
                new_bound_keys, dt["_hilbert_key_grid"][mask], side="left"
            )
            new_cpu_map = (
                np.searchsorted(new_bound_keys, dt["_hilbert_key"], side="left")
                .astype(np.int32)
                .reshape(-1, 8)
            )

        dt["_new_cpu_map"] = new_cpu_map

        lvl_ind = dt["_level"][mask].astype(np.int64)

        owing_cpu = np.full_like(new_domain_ind, icpu)
        N2 += oct.add(
            ipos,
            file_ind,
            domain_ind,
            new_domain_ind,
            owing_cpu,
            dt["_hilbert_key"].reshape(-1, 8),
            dt["refmap"],
            lvl_ind,
        )

    print(f"Inserted {N2} octs")

    dt = data[1]
    igrid0 = dt["ind_grid"][1:]
    icell, igrid = np.unravel_index(dt["parent"][igrid0 - 1] - ncoarse, (8, ngridmax))

    np.testing.assert_allclose(
        (((dt["xc"][igrid - 1] - dt["xc"][1:]) * 2 ** dt["_level"][1:, None]) ** 2).sum(
            axis=1
        ),
        3,
    )

    def set_neighbours(dt):
        lvl = dt["_level"].astype(int)[1:]
        igrid0 = dt["ind_grid"][1:]

        icell_neigh, igrid_neigh = np.unravel_index(
            dt["nbor"][igrid0 - 1] - ncoarse, (8, ngridmax)
        )
        dd = 1 / 2 ** dt["_level"][igrid_neigh - 1][..., None] / 2
        dx = np.array([-1, 1, -1, 1, -1, 1, -1, 1])
        dy = np.array([-1, -1, 1, 1, -1, -1, 1, 1])
        dz = np.array([-1, -1, -1, -1, 1, 1, 1, 1])
        xc = dt["xc"]
        xc_neigh = xc[igrid_neigh - 1] + dd * np.stack(
            [dx[icell_neigh], dy[icell_neigh], dz[icell_neigh]], axis=-1
        )

        ixc = (xc[igrid0 - 1] * bscale).astype(int).copy()
        ixc_neigh = (xc_neigh * bscale).astype(int).copy()

        # Set neighbours for all octs that do have a child
        oct.set_neighbours(ixc, ixc_neigh, lvl)

    for dt in tqdm(data.values(), desc="Setting neighbours"):
        set_neighbours(dt)

    # oct.print_tree(3, print_neighbours=True)

    QUADHILBERT = "float128" if args.quadhilbert else "float64"
    LONGINT = "int64" if args.longint else "int32"

    _HEADERS = (
        ("ncpu", "i"),
        ("ndim", "i"),
        ("nx", "i"),
        ("nlevelmax", "i"),
        ("ngridmax", "i"),
        ("nboundary", "i"),
        ("ngrid_current", "i"),
        ("boxlen", "d"),
        ("nout", "i"),
        ("tout", "d"),
        ("aout", "d"),
        ("t", "d"),
        ("dtold", "d"),
        ("dtnew", "d"),
        ("nstep", "i"),
        ("stat", "d"),
        ("cosm", "d"),
        ("timing", "d"),
        ("mass_sph", "d"),
    )

    _AMR_STRUCT = (
        ("headl", "i"),
        ("taill", "i"),
        ("numbl", "i"),
        ("numbtot", LONGINT),
        # BOUNDARIES: not supported
        (("headf", "tailf", "numbf", "used_mem", "used_mem_tot"), "i"),
        ("ordering", 128),  # 128-long char array
        ("bound_keys", QUADHILBERT),
    )

    def write_amr_file(headers, amr_struct, amr_file):
        """This write the new amr files

        Parameters
        ----------
        """

        # Make sure arrays have the right types
        def convert(key, dtype):
            amr_struct[key] = amr_struct[key].astype(dtype)

        for key in (
            "headl",
            "numbl",
            "ind_grid",
            "next",
            "prev",
            "parent",
            "nbor",
            "son",
            "cpu_map",
            "refmap",
        ):
            convert(key, np.int32)
        for key in ("xc",):
            convert(key, np.float64)
        # Write headers
        f = FF(amr_file, mode="w")
        print("Writing headers. ", end="")
        for k, t in _HEADERS:
            tmp = np.atleast_1d(headers[k]).astype(t)
            print(k, end="...")
            f.write_vector(tmp)
        print()

        # Write global AMR structure
        print("Writing global AMR structure. ", end="")
        for k, t in _AMR_STRUCT:
            if not isinstance(t, int):
                tmp = np.ascontiguousarray(np.atleast_1d(amr_struct[k]), dtype=t)
            else:
                tmp = np.char.asarray(amr_struct[k].ljust(128).encode(), 1).astype("c")
            print(f"{k}", end="...")
            f.write_vector(tmp)

        # Coarse level
        numbl = amr_struct["numbl"]
        print("Coarse level")
        # NOTE: since son has shape (ncoarse + ngridmax*8, we only need to write cell 0
        #       of coarse level (i.e. root oct)
        f.write_vector(amr_struct["coarse_son"].astype(np.int32))
        f.write_vector(amr_struct["coarse_refmap"].astype(np.int32))
        cpu_map_coarse = np.argwhere(numbl[0] == 1).astype(np.int32).flatten() + 1
        assert cpu_map_coarse.size == 1
        f.write_vector(cpu_map_coarse)

        print("Fine levels")
        nlevelmax = headers["nlevelmax"]
        ncpu = headers["ncpu"]
        nboundary = headers["nboundary"]
        if nboundary > 0:
            raise NotImplementedError

        ii = 0
        ncache = 0

        def write_chunk(key, extra_slice=...):
            f.write_vector(
                np.ascontiguousarray(amr_struct[key][ii : ii + ncache, extra_slice])
            )

        for ilvl in range(nlevelmax):
            for ibound in range(ncpu + nboundary):
                if ibound < ncpu:
                    ncache = numbl[ilvl, ibound]
                # else:
                #     ncache = numbb[ilvl, ibound]
                if ncache == 0:
                    continue
                write_chunk("ind_grid")
                write_chunk("next")
                write_chunk("prev")
                for idim in range(3):
                    write_chunk("xc", idim)
                write_chunk("parent")
                for idim in range(2 * 3):
                    write_chunk("nbor", idim)
                for idim in range(2 ** 3):
                    write_chunk("son", idim)
                for idim in range(2 ** 3):
                    assert np.all(amr_struct["cpu_map"][ii : ii + ncache, idim] > 0)
                    write_chunk("cpu_map", idim)
                for idim in range(2 ** 3):
                    write_chunk("refmap", idim)

                ii += ncache

    Noct_tot = oct.count_octs()

    # Create AMR structure
    new_data = {}
    for new_icpu in range(1, CONFIG["new_ncpu"] + 1):
        bk_low = new_bound_keys[new_icpu - 1]
        bk_up = new_bound_keys[new_icpu]

        print(
            f'Selecting grid intersecting with new cpu #{new_icpu}/{CONFIG["new_ncpu"]}'
        )

        ###########################################################
        # Headers
        dt = data[1]
        headers = dt["headers"].copy()
        headers["ncpu"] = CONFIG["new_ncpu"]

        ###########################################################
        # Amr structure
        oct.clear_paint()
        if args.remap == "at_boundaries":
            old_icpu_list = list(
                range(1 + icpu_ind[new_icpu - 1], 1 + icpu_ind[new_icpu])
            )
            print(new_icpu, old_icpu_list)

            if len(old_icpu_list) == 0:
                dt = data[icpu_ind[new_icpu - 1] + 1]
                all_lvl = dt["_level"].astype(np.int64)[1:]
                mask = all_lvl <= ds.min_level + 1
                lvl = all_lvl[mask]
                xc = dt["xc"]
                igrid0 = dt["ind_grid"][1:][mask]

                ixc = (xc[igrid0 - 1] * bscale).astype(np.int64).copy()

                oct.select(ixc, lvl)

            # Loop over old domains and select the cells in them
            for dt in (data[_] for _ in old_icpu_list):
                xc = dt["xc"]
                lvl = dt["_level"].astype(np.int64)[1:]
                igrid0 = dt["ind_grid"][1:]
                ixc = (xc[igrid0 - 1] * bscale).astype(np.int64).copy()

                oct.select(ixc, lvl)
        else:
            # Select cells that intersect with domain
            oct.select_hilbert(bk_low, bk_up)
            oct.expand_boundaries(new_icpu, nexpand=args.nexpand)

        # Extract structure
        amr_struct = oct.extract_data()

        file_inds = amr_struct["file_ind"]
        nocts = len(file_inds)

        amr_struct["bound_keys"] = new_bound_keys
        amr_struct["numbtot"] = dt["numbtot"]
        amr_struct[("headf", "tailf", "numbf", "used_mem", "used_mem_tot")] = (
            0,
            0,
            0,
            0,
            0,
        )
        amr_struct["ordering"] = "hilbert"
        amr_struct["coarse_son"] = dt["coarse_son"]
        amr_struct["coarse_refmap"] = dt["coarse_refmap"]

        if args.ngridmax == 0:
            headers["ngridmax"] = Noct_tot * 2 // args.ncpu
        elif args.ngridmax == -1:
            headers["ngridmax"] = headers["ngridmax"]
        else:
            headers["ngridmax"] = args.ngridmax

        def pair2icell(v):
            # Compute cell indices from parent oct + icell
            return v[..., 0] + ncoarse + v[..., 1] * headers["ngridmax"]

        amr_struct["parent"] = pair2icell(amr_struct["parent"])
        amr_struct["nbor"] = pair2icell(amr_struct["nbor"])
        amr_struct["cpu_map"] = amr_struct["new_domain_ind"]
        amr_struct["ind_grid"] = np.arange(1, nocts + 1, dtype=np.int32)
        amr_struct["headers"] = headers
        new_data[new_icpu] = amr_struct

        # Compute refmap, etc.
        # fields = (('refmap', 'refmap'), ('xc', 'xc'), ('cpu_map', '_new_cpu_map'))
        fields = (("xc", "xc"), ("cpu_map", "_new_cpu_map"))

        data_out = {}

        dt = data[1]
        for f_new, f_old in fields:
            data_out[f_new] = np.zeros(
                [nocts] + list(dt[f_old].shape[1:]), dtype=dt[f_old].dtype
            )

        for icpu, dt in data.items():
            mask = amr_struct["owning_cpu"] == icpu
            find = file_inds[mask] - 1
            for f_new, f_old in fields:
                data_out[f_new][mask] = dt[f_old][find]

        for k, v in data_out.items():
            amr_struct[k] = v

        # Make sure ngridmax is large enough
        if nocts > headers["ngridmax"]:
            raise RuntimeError(
                "ERROR: you need to increase ngridmax to at least %s, got %s!"
                % (nocts, headers["ngridmax"])
            )

        # Write AMR file
        base = args.output_dir
        os.makedirs(base, exist_ok=True)
        iout = CONFIG["iout"]
        amr_file = os.path.join(base, f"amr_{iout:05d}.out{new_icpu:05d}")

        tmp = amr_struct
        write_amr_file(headers, tmp, amr_file)

    print(" old structure ".center(120, "="))
    for l in data[1]["numbl"]:
        for c in l:
            print("%7d" % c, end="")
        print("| %s" % l.sum())

    print(" new structure ".center(120, "="))

    for l in new_data[1]["numbl"]:
        for c in l:
            print("%7d" % c, end="")
        print("| %s" % l.sum())

    oct.check_tree(999)

    # Now write hydro

    def fluid_file_reader(field_handler: FieldFileHandler, headers: dict = None):
        if headers is None:
            headers = {}
        with FF(field_handler.fname, "r") as fin:
            headers.update(fin.read_attrs(field_handler.attrs))
            nvar = headers["nvar"]
            nboundaries = headers["nboundary"]
            nlevelmax = headers["nlevelmax"]
            ncpu = headers["ncpu"]

            data_out = np.full(
                (nvar, 8, field_handler.domain.amr_header["ngrid_current"]),
                0,
                dtype="d",
            )

            ii = 0
            for _ilvl in range(nlevelmax):
                for _icpu in range(ncpu + nboundaries):
                    fin.read_int()  # ilvl2
                    ncache = fin.read_int()
                    if ncache > 0:
                        for icell in range(8):
                            for ivar in range(nvar):
                                data_out[
                                    ivar, icell, ii : ii + ncache
                                ] = fin.read_vector("d")
                    ii += ncache
        return headers, data_out

    def fluid_file_writer(fname, headers, data, numbl):
        with FF(fname, mode="w") as fout:
            for key, _len, dtype in headers["_structure"]:
                tmp = np.atleast_1d(headers[key]).astype(dtype)
                assert len(tmp) == _len
                fout.write_vector(tmp)
            nvar = headers["nvar"]
            nboundaries = headers["nboundary"]
            nlevelmax = headers["nlevelmax"]
            ncpu = headers["ncpu"]

            ii = 0

            for ilvl in range(1, 1 + nlevelmax):
                for icpu in range(1, ncpu + nboundaries + 1):
                    ncache = numbl[ilvl - 1, icpu - 1]
                    fout.write_vector(np.asarray([ilvl], dtype=np.int32))
                    fout.write_vector(np.asarray([ncache], dtype=np.int32))
                    if ncache > 0:
                        for icell in range(8):
                            for ivar in range(nvar):
                                tmp = data[ivar, icell, ii : ii + ncache]
                                fout.write_vector(np.ascontiguousarray(tmp))
                    ii += ncache

    def write_fluid_file(filename, amr_structure, headers, data_old):
        ncpu_old = len(data_old)

        # Loop over AMR structure, and read relevant files
        nvar = headers["nvar"]
        nocts = amr_structure["owning_cpu"].size
        data_new = np.full((nvar, 8, nocts), np.nan, dtype="d")

        for icpu_old in range(1, 1 + ncpu_old):
            mask = amr_structure["owning_cpu"] == icpu_old  # select octs in this file
            n_to_read = mask.sum()
            if n_to_read == 0:
                continue

            # Original positions
            file_ind = amr_structure["file_ind"][mask] - 1

            # Target position
            new_file_ind = amr_structure["ind_grid"][mask] - 1
            data_new[..., new_file_ind] = data_old[icpu_old][..., file_ind]

        # Write file
        fluid_file_writer(filename, headers, data_new, amr_structure["numbl"])

    class FluidFileAttrs:
        fname_pattern = None

    class GravityFluidFileAttrs(FluidFileAttrs, GravFieldFileHandler):
        fname_pattern = "grav_{iout:05d}.out{icpu:05d}"
        ftype = None  # prevent yt from using this to detect files

    class HydroFluidFileAttrs(FluidFileAttrs, HydroFieldFileHandler):
        fname_pattern = "hydro_{iout:05d}.out{icpu:05d}"
        ftype = None  # prevent yt from using this to detect files

    fluid_descs = {"gravity": GravityFluidFileAttrs, "ramses": HydroFluidFileAttrs}

    def rewrite_fluid_files(amr_structure, domains, output_dir, iout):
        nkind = len(domains[0].field_handlers)
        ncpu_new = len(amr_structure)
        all_fdescs = [fluid_descs[fh.ftype] for fh in domains[0].field_handlers]

        progress = tqdm(total=nkind)
        for i, fdesc in enumerate(all_fdescs):
            ftype = domains[0].field_handlers[i].ftype
            progress.set_description(f"{ftype}: R")
            data_orig = {}
            headers = {}
            for icpu, dom in enumerate(
                tqdm(domains, desc="Reading files", leave=False)
            ):
                fh = dom.field_handlers[i]
                ret = fluid_file_reader(fh, default_headers[ftype].copy())
                headers, data_orig[icpu + 1] = ret

            # Need to update the number of cpus in the headers
            headers["ncpu"] = ncpu_new

            headers["_structure"] = fdesc.attrs

            progress.set_description(f"{ftype}: W")
            for icpu in tqdm(amr_structure.keys(), desc="Writing files", leave=False):
                fname = os.path.join(
                    output_dir, fdesc.fname_pattern.format(iout=iout, icpu=icpu)
                )
                write_fluid_file(
                    fname, amr_structure[icpu], headers, data_old=data_orig
                )
            progress.update(1)

    rewrite_fluid_files(
        new_data, ds.index.domains, output_dir=args.output_dir, iout=CONFIG["iout"]
    )

    class ParticleFileAttrs(DefaultParticleFileHandler):
        fname_pattern = "part_{iout:05d}.out{icpu:05d}"
        ptype = None  # prevent yt from using this to detect files

    class SinkFileAttrs(SinkParticleFileHandler):
        fname_pattern = "sink_{iout:05d}.out{icpu:05d}"
        ptype = None

    particle_descs = {"io": ParticleFileAttrs, "sink": SinkFileAttrs}

    def particle_file_reader(
        particle_handler: ParticleFileHandler, headers: dict = None
    ) -> dict:
        if headers is None:
            headers = {}

        data_out = {}
        with FF(particle_handler.fname, "r") as fin:
            headers.update(fin.read_attrs(particle_handler.attrs))
            try:
                npart = headers["npart"]
            except KeyError:
                npart = headers[""]
            npart = headers.get("npart", None)

            for k, dtype in particle_handler.field_types.items():
                data_out[k] = fin.read_vector(dtype)

                # This happens for sink files
                if npart is not None:
                    assert data_out[k].size == npart

        return headers, data_out

    def particle_file_writer(fname, particle_new_domain, headers, data_old, new_icpu):
        # Count number of particles
        npart = 0
        masks = {}
        for icpu_old, part_dom in particle_new_domain.items():
            mask = part_dom == new_icpu
            masks[icpu_old] = mask

            npart += mask.sum()

        # Extract fields
        fields = data_old[1].keys()

        with FF(fname, mode="w") as fout:
            h = headers.copy()
            h["npart"] = npart

            # Write headers
            for key, _len, dtype in h["_structure"]:
                tmp = np.atleast_1d(h[key]).astype(dtype)
                assert (len(tmp) == _len) or (_len == -1)
                fout.write_vector(tmp)

            # Write fields
            for field in fields:
                vals = np.empty(npart, dtype=data_old[1][field].dtype)

                i0 = 0
                for mask, dt in zip(masks.values(), data_old.values()):
                    count = mask.sum()
                    vals[i0 : i0 + count] = dt[field][mask]
                    i0 += count

                fout.write_vector(vals)

    def rewrite_particle_files(amr_structure, domains, output_dir, iout):
        ncpu_new = len(amr_structure)

        particle_handlers = domains[0].particle_handlers
        nkind = len(particle_handlers)

        progress = tqdm(total=nkind)

        # Special case for sinks -> simply make new_ncpu copies (all sink files are the same)
        for ph in (ph for ph in particle_handlers if ph.ptype == "sink"):
            progress.set_description(f"{ph.ptype}: R")
            pdesc = particle_descs[ph.ptype]
            original_fname = os.path.join(
                input_dir, pdesc.fname_pattern.format(iout=iout, icpu=1)
            )
            for icpu in tqdm(
                amr_structure.keys(), desc="Writing sink files", leave=False
            ):
                fname = os.path.join(
                    output_dir, pdesc.fname_pattern.format(iout=iout, icpu=icpu)
                )
                shutil.copy(original_fname, fname)
            progress.update()

        new_bound_keys = amr_structure[1]["bound_keys"]

        for i, ph in enumerate(particle_handlers):
            if ph.ptype == "sink":
                continue
            progress.set_description(f"{ph.ptype}: R")
            pdesc = particle_descs[ph.ptype]
            data_orig = {}
            particle_new_domain = {}
            headers = {}
            for icpu, dom in enumerate(
                tqdm(domains, desc="Reading files", leave=False)
            ):
                fh = dom.particle_handlers[i]
                ret = particle_file_reader(fh, {})
                headers, data_orig[icpu + 1] = ret
                pos = np.stack(
                    [ret[1]["io", "particle_position_%s" % k] for k in "xyz"], axis=-1
                )
                ipos = np.round(pos * bscale).astype(np.int64)
                particle_new_domain[icpu + 1] = np.searchsorted(
                    new_bound_keys,
                    hilbert3d(ipos, bit_length),
                    side="left",
                )

            # Need to update the number of cpus in the headers
            headers["ncpu"] = ncpu_new
            headers["_structure"] = pdesc.attrs

            progress.set_description(f"{ph.ptype}: W")
            for icpu in tqdm(amr_structure.keys(), desc="Writing files", leave=False):
                fname = os.path.join(
                    output_dir, pdesc.fname_pattern.format(iout=iout, icpu=icpu)
                )
                particle_file_writer(
                    fname,
                    particle_new_domain,
                    headers,
                    data_old=data_orig,
                    new_icpu=icpu,
                )
            progress.update()

    rewrite_particle_files(
        new_data, ds.index.domains, output_dir=args.output_dir, iout=CONFIG["iout"]
    )

    # Final touch: copy namelist, file descriptors and info
    def copy_meta(input_info_file, output_dir, bound_key, new_ncpu):
        input_dir, input_info_fname = os.path.split(input_info_file)
        output_info_file = os.path.join(output_dir, input_info_fname)

        # Rewrite "info_XXXXX.txt"
        with open(input_info_file) as fin:
            lines = fin.readlines()
        with open(output_info_file, "w") as fout:
            line = lines.pop(0)
            while not line.strip().startswith("DOMAIN   ind_min"):
                if "ncpu" in line:
                    line = "ncpu        = %10d\n" % new_ncpu
                fout.write(line)
                line = lines.pop(0)

            # Write hilbert keys
            fout.write("   DOMAIN   ind_min                 ind_max\n")
            for icpu in range(1, new_ncpu + 1):
                s = f"{icpu:8d}   {bound_key[icpu-1]:.15E}   {bound_key[icpu]:.15E}\n"
                fout.write(s)

        # Copy file descriptors, header & namelist
        tgt_files = glob(os.path.join(input_dir, "*file_descriptor.txt"))
        tgt_files += glob(os.path.join(input_dir, "header_?????.txt"))
        tgt_files += glob(os.path.join(input_dir, "namelist.txt"))
        for f_in in tgt_files:
            f_out = os.path.join(output_dir, os.path.split(f_in)[1])
            shutil.copy(f_in, f_out)

    copy_meta(args.input, args.output_dir, new_bound_keys, CONFIG["new_ncpu"])

    if args.test:
        print("Testing using yt.".center(200, "="))
        test_using_yt(args, CONFIG)


if __name__ == "__main__":
    main()


def test_using_yt(args, CONFIG):
    yt.funcs.mylog.setLevel(40)
    # Test reading with yt as a weak test
    ds_original = yt.load(args.input)
    ds_new = yt.load(os.path.join(args.output_dir, "info_%05d.txt" % CONFIG["iout"]))

    images = []
    os.makedirs("tmp/frames/", exist_ok=True)
    for ds, prefix in reversed(
        list(zip((ds_new, ds_original, ds_new), ("new", "ref")))
    ):
        print(f"Loading = {ds}")
        for d in "xyz":
            p = yt.ProjectionPlot(ds, d, "density")
            images.extend(p.save(f"tmp/frames/{d}_{prefix}"))

    for ds, prefix in zip((ds_original, ds_new), ("ref", "new")):
        for d in "xyz":
            p = yt.ProjectionPlot(ds, d, "DM_cic")
            images.extend(p.save(f"tmp/frames/{d}_{prefix}"))
    print("You can compare the images look the same using the images from %s" % images)

    # %%
    images_with_diff = list(images)

    def raw(i):
        return mpl.image.imread(images[i])

    for i in range(3):
        new_fname = images[i].replace("_ref", "_diff")
        diff = (raw(3 + i) - raw(i) + 1) / 2
        mpl.image.imsave(new_fname, diff)
        images_with_diff.append(new_fname)

    for i in range(6, 6 + 3):
        new_fname = images[i].replace("_ref", "_diff")
        diff = (raw(3 + i) - raw(i) + 1) / 2
        mpl.image.imsave(new_fname, diff)
        images_with_diff.append(new_fname)

    print(
        "Note: we expect artifacts with particle CIC deposition due to yt internal deposition (does not depose on other domains)"
    )
    ipyplot.plot_class_tabs(
        np.asarray(images_with_diff),
        np.asarray(["_".join(_.split("_")[-3:]) for _ in images_with_diff]),
        img_width=400,
    )

    # Compare gas cells
    order = {}

    ad_new = ds_new.all_data()
    ad_ref = ds_original.all_data()

    for ad, prefix in reversed(list(zip((ad_new, ad_ref), ("new", "ref")))):
        order[prefix] = np.argsort(ad["index", "morton_index"])

    def extract_field(ad, field_name, order):
        return ad[field_name][order]

    for field in (
        "Density Pressure Metallicity".split()
        + [f"{k}-velocity" for k in "xyz"]
        + [_ for _ in "xyz"]
    ):
        print("Checking %s" % field, end="... ")
        vnew = extract_field(ad_new, field, order["new"])
        vref = extract_field(ad_ref, field, order["ref"])

        np.testing.assert_allclose(vnew, vref)
        print("ok!")

    # Compare gas cells
    order = {}

    ad_new = ds_new.all_data()
    ad_ref = ds_original.all_data()

    for ad, prefix in reversed(list(zip((ad_new, ad_ref), ("new", "ref")))):
        order[prefix] = np.argsort(ad["io", "particle_identity"])

    def extract_field(ad, field_name, order):
        return ad[field_name][order]

    for field in (
        ("DM", f"particle_{suffix}")
        for suffix in "mass position velocity family tag".split()
    ):
        print("Checking %s, %s" % field, end="... ")
        vnew = extract_field(ad_new, field, order["new"])
        vref = extract_field(ad_ref, field, order["ref"])

        np.testing.assert_allclose(vnew, vref)
        print("ok!")


def debug_grid(dt, ncoarse):
    # ## Debugging the grid
    def print_tree(ioct, dt):
        header = "     ioct  xc                                     cpu_map                    son                                                cpu_neigh            neigh_ind                            "
        print(header)
        print("—" * len(header))
        while ioct >= 1:
            __ = lambda e: e[e > 0]  # noqa: E731
            nbor = dt["nbor"][ioct - 1]
            mask = nbor > 0
            neigh_icell, neigh_igrid = (np.full(6, -1, dtype=int) for _ in range(2))
            neigh_icell[mask], neigh_igrid[mask] = np.unravel_index(
                nbor[mask] - ncoarse, (8, ngridmax)
            )
            cpu_neigh = dt["cpu_map"][neigh_igrid - 1, neigh_icell]
            cpu_neigh[~mask] = -1

            def __(x, n):
                return str(x).ljust(n)

            neigh_ind = dt["son"][neigh_igrid - 1, neigh_icell]
            neigh_ind[~mask] = -1
            print(
                "",
                f"{ioct:>8d}",
                __(dt["xc"][ioct - 1], 29),
                "\t",
                __(dt["cpu_map"][ioct - 1], 26),
                __(dt["son"][ioct - 1], 50),
                __(cpu_neigh, 20),
                __(neigh_ind, 20),
            )
            ioct = (dt["parent"][ioct - 1] - ncoarse) % ngridmax

    DIRECTIONS = [0, 1, 2]

    def get_cube(ioct, dt, ret=None, prev_directions=None, depth=0, path=""):
        if prev_directions is None:
            prev_directions = []
        if ret is None:
            ret = {}
            ret["iocts"] = {ioct}
            ret["path"] = defaultdict(list)
            ret["path"][ioct] = [""]

        for idir in (_ for _ in DIRECTIONS if _ not in prev_directions):
            __ = lambda e: e[e > 0]  # noqa: E731
            neigh_icell, neigh_igrid = np.unravel_index(
                __(dt["nbor"][ioct - 1]) - ncoarse, (8, ngridmax)
            )
            iocts_neigh = dt["son"][neigh_igrid - 1, neigh_icell][
                2 * idir : 2 * idir + 2
            ]
            for ii, ioct_neigh in enumerate(iocts_neigh):
                if ioct_neigh == 0:
                    continue
                new_directions = prev_directions + [idir]
                ret["iocts"].add(ioct_neigh)
                new_path = str(path)
                new_path += "-+"[ii] + "xyz"[idir]
                ret["path"][ioct_neigh].append(new_path)
                get_cube(
                    ioct_neigh,
                    dt,
                    ret=ret,
                    prev_directions=new_directions,
                    depth=depth + 1,
                    path=new_path,
                )

        paths = ret["path"]
        iocts = sorted(ret["iocts"], key=lambda k: (len(paths[k][0]), paths[k][0]))
        return iocts, {ioct: paths[ioct] for ioct in iocts}

    def match_tree(ioct, dt, new_dt):
        """Find an oct in the other tree."""
        # Walk the tree up
        icell_list = []
        while ioct > 1:
            __ = lambda e: e[e > 0]  # noqa: E731
            _neigh_icell, _neigh_igrid = np.unravel_index(
                __(dt["nbor"][ioct - 1]) - ncoarse, (8, ngridmax)
            )
            ioct_parent = (dt["parent"][ioct - 1] - ncoarse) % ngridmax
            my_index = np.argwhere(dt["son"][ioct_parent - 1] == ioct).flatten()[0]
            icell_list.append(my_index)
            ioct = ioct_parent

        ioct_old = 1
        ioct = 1
        ioct_prev = -1
        for icell in reversed(icell_list):
            ioct_prev = ioct
            ioct = new_dt["son"][ioct - 1][icell]
            ioct_old = dt["son"][ioct_old - 1][icell]
            print(f"{ioct_old:10d} {ioct:10d}")
        return ioct_prev

    ngridmax = dt["headers"]["ngridmax"]
    icell, iparent = np.unravel_index(dt["parent"] - ncoarse, (8, ngridmax))
    iparent += 1

    assert np.all(iparent[1:] > 0)

    def get_this_neighbour(i, j, k, dt, ioct):
        icell, igrid = np.unravel_index(
            dt["nbor"][ioct - 1, i] - ncoarse, (8, ngridmax)
        )
        ioct = dt["son"][igrid, icell]

        icell, igrid = np.unravel_index(
            dt["nbor"][ioct - 1, j] - ncoarse, (8, ngridmax)
        )
        ioct = dt["son"][igrid, icell]

        icell, igrid = np.unravel_index(
            dt["nbor"][ioct - 1, k] - ncoarse, (8, ngridmax)
        )
        ioct = dt["son"][igrid, icell]

        return ioct

    for i in range(6):
        for j in range(i + 1, 6):
            if i // 2 == j // 2:
                continue
            for k in range(j + 1, 6):
                if k // 2 == j // 2:
                    continue
                break
            break
        break

    def check_domain(dt):
        ngridmax = dt["headers"]["ngridmax"]
        print("\t\tChecking neighbours")
        icell, igrid = np.unravel_index(dt["nbor"] - ncoarse, (8, ngridmax))
        icell, igrid

        # Make sure all neighbours exist
        assert np.all(igrid[1:] > 0)

        # Make sure all neighbours' neighbour exist
        for ioct in dt["ind_grid"][9:]:
            son_mask = dt["son"][ioct - 1] > 0
            if son_mask.all():
                cube = get_cube(ioct, dt)[0]
                if len(cube) != 27:
                    raise Exception(f"{ioct} is missing {27-len(cube)} neighbours")

    def check_tree(data):
        # Check that each grid has its parent neighbour grid
        for icpu, dt in data.items():
            print(f"\tChecking domain {icpu}")
            check_domain(dt)

    # # %%
    # print("Original")
    # check_tree(data)
    # print("Rewritten")
    # check_tree(new_data)



import numpy as np
import pytest

from ..periodicity import distance_to_center, wrap_coordinates


@pytest.fixture
def seed():
    return 16091992


def test_wrap_coordinates(seed):
    np.random.seed(seed)
    x = np.random.rand(1000, 3) * 3 - 1

    xw = wrap_coordinates(x)
    assert np.all(xw >= 0)
    assert np.all(xw < 1)

    xw = wrap_coordinates(x, 0.5)
    assert np.all(xw >= 0)
    assert np.all(xw < 0.5)


def test_distance_to_center(seed):
    np.random.seed(seed)
    x = np.random.rand(1000, 3) * 3 - 1

    dx = distance_to_center(x)
    assert np.all(dx >= -0.5)
    assert np.all(dx <= 0.5)

    dx = distance_to_center(x, 0.5)
    assert np.all(dx >= -0.25)
    assert np.all(dx <= 0.25)

    x = np.random.normal(loc=0.5, scale=0.001, size=(10_000, 3))
    dx_ref = x - x.mean(axis=0)
    dx = distance_to_center(x)

    np.testing.assert_allclose(dx, dx_ref, atol=1e-7)



from shutil import which

import numpy as np
import pytest

from ..yorick import PDBReader

SKIP_YORICK = which("yorick") is None
SKIP_YORICK_REASON = "Could not find yorick executable. Skipping tests."


@pytest.fixture
def pdb():
    return PDBReader("data/galaxies-040.pdb")


def walker_helper(keys, root):
    for key, val in root.items():
        if isinstance(val, dict):
            yield from walker_helper(keys + [key], val)
        else:
            if isinstance(val, tuple):
                exp_type = val[0]
            else:
                exp_type = val
            yield keys + [key], exp_type


def deep_dict_compare(a, b, prefix=None, print_pre=""):
    if prefix is None:
        prefix = []
    assert len(a) == len(b)

    key_a = set(a.keys())
    key_b = set(b.keys())

    if len(key_a.difference(key_b)) > 0:
        raise AssertionError("Found different set of keys at %s" % prefix)

    for key, va in a.items():
        print(print_pre, end="")
        vb = b[key]
        if type(va) != type(vb):
            raise AssertionError(f"Type differ at {prefix}[{key}]")

        if isinstance(va, dict):
            print("checking key «%s»" % key)
            deep_dict_compare(va, vb, prefix + [key], print_pre + "\t")
        else:
            print(f"comparing {prefix} {key}", end="...")
            if va != vb:
                raise AssertionError(f"Value differ at {prefix}[{key}], {va} ≠ {vb}")
            print("ok!")


@pytest.mark.skipif(SKIP_YORICK, reason=SKIP_YORICK_REASON)
def test_structure(pdb):
    expected_structure = {
        "catalog": {
            "vir": {"rvir": float, "mvir": float, "tvir": float, "cvel": float},
            "shape": (float, 3),
            "pos": (float, 3),
            "vel": (float, 3),
            "L": (float, 3),
            "profile": (float, 2),
            "age": float,
            "aexp": float,
            "num": int,
            "slice": (int, 2),
            "level": int,
            "hosthalo": int,
            "hostsub": int,
            "nbsub": int,
            "nextsub": int,
            "mass": float,
            "rad": float,
            "spin": float,
            "ek": float,
            "ep": float,
            "et": float,
            "macc": float,
            "tree": {
                "nbfather": int,
                "father": "yorick_pointer",
                "mfrac": "yorick_pointer",
                "nbson": int,
                "son": "yorick_pointer",
            },
            "npart": int,
            "index": "yorick_pointer",
            "bulge": (float, 3),
        }
    }
    deep_dict_compare(pdb.structure, expected_structure)


@pytest.mark.skipif(SKIP_YORICK, reason=SKIP_YORICK_REASON)
def test_access(pdb):
    # This should work
    for keys, expected_type in walker_helper([], pdb.structure):
        # Bake access path as "a/b/c"
        path = "/".join(keys)
        res = pdb[path]

        if expected_type in (int, float):
            assert isinstance(res, np.ndarray)


@pytest.mark.skipif(SKIP_YORICK, reason=SKIP_YORICK_REASON)
def test_access_incorrect(pdb):
    with pytest.raises(KeyError):
        pdb["this/does/not/exists"]



cimport cython
cimport numpy as np


cdef extern from *:
    ctypedef int int128_t "__int128_t"
    ctypedef float float128_t "__float128"

cdef np.int64_t interleave_3bits_64(np.int64_t x) nogil
cdef np.uint64_t hilbert3d_single(const np.int64_t X[3], const int bit_length) nogil
cdef np.uint64_t[:] hilbert3d_int(np.int64_t[:, ::1] X, int bit_length)

cdef int128_t interleave_3bits_128(int128_t x) nogil
cdef int128_t hilbert3d_single_quad(const np.int64_t X[3], const int bit_length) nogil



import numpy as np
from numba import float32, float64, vectorize


@vectorize([float64(float64), float32(float32)])
def j1_over_x(x):
    """Compute j_1(x)/x.

    Parameter
    ---------
    x : float, or array-like

    Return
    ------
    j_1(x) / x : float, or array-like
    """
    x2 = x * x
    if x < 1e-2:
        return (1 - x2 * 0.1 * (1 - x2 / 28)) / 3
    # elif x < 8e-2:
    #     return (1 - x2*0.1*(1 - x2/28*(1 - x2/54.)))/3
    elif x < 2e-1:
        return (1 - x2 * 0.1 * (1 - x2 / 28 * (1 - x2 / 54.0 * (1 - x2 / 88.0)))) / 3
    else:
        return (np.sin(x) / x - np.cos(x)) / x2



"""Support Disperse I/O.

The reader is based on a description of the structure using Kaitai (https://kaitai.io/).
"""

import numpy as np
import pandas as pd

from ..utilities.decorators import read_files
from ..utilities.types import FloatArrayType, PathType
from .disperse_reader import DisperseReader


class Disperse:
    """Read a disperse NDskl file.

    Parameters
    ----------
    fname : str, filename
    """

    @read_files(1)
    def __init__(self, fname: PathType):
        self.fname = fname
        self.read()

    def read(self):
        ds = DisperseReader.from_file(self.fname)

        # Gather data
        nnode, nnode_data, ndim = ds.header.nnode, ds.header.nnode_data, ds.header.ndim
        node_pos = np.asarray(ds.data.node_pos).reshape((nnode, ndim))
        node_data = np.asarray(ds.data.node_data).reshape((nnode, nnode_data))

        nseg, nseg_data = ds.header.nseg, ds.header.nseg_data
        seg_data = np.asarray(ds.data.seg_data).reshape((nseg, nseg_data))

        #  # Gather information about nodes
        _v1 = pd.DataFrame(
            [
                (*node_pos[s0.pos_index], s0.index, s0.flags, s0.type)
                for s0 in ds.data.node_data_int
            ],
            columns=["x", "y", "z", "index", "flags", "type"],
        ).set_index("index")

        _v2 = pd.DataFrame(
            node_data,
            columns=[_.replace("\x00", "") for _ in ds.header.node_data_info],
            index=_v1.index,
        )
        # Parse what's integer as integer
        int_columns = ["parent_index", "persistence_pair"]
        for c in int_columns:
            _v2[c] = _v2[c].astype(int)

        node_ds = pd.concat((_v1, _v2), axis=1)
        node_ds["type_s"] = [
            ("void", "wall", "filament", "peak")[_] for _ in node_ds.type
        ]

        #  # Gather information about segments
        _v1 = pd.DataFrame(
            [
                (s0.index, s0.node_ids[0], s0.node_ids[1], s0.prev_seg, s0.next_seg)
                for s0 in ds.data.seg_data_int
            ],
            columns=["index", "node_start", "node_end", "seg_prev", "seg_next"],
        )
        _v1 = _v1.set_index("index")
        _v2 = pd.DataFrame(
            seg_data,
            columns=[_.replace("\x00", "") for _ in ds.header.seg_data_info],
            index=_v1.index,
        )
        # Parse what's integer as integer
        int_columns = ["type", "orientation"]
        for c in int_columns:
            _v2[c] = _v2[c].astype(int)
        # Replace off-bound values by dummy values
        _v1.loc[_v1["seg_prev"] > ds.header.nseg, "seg_prev"] = ds.header.nseg
        _v1.loc[_v1["seg_next"] > ds.header.nseg, "seg_next"] = ds.header.nseg

        seg_ds = pd.concat((_v1, _v2), axis=1)

        seg_ds["node_start_type"] = node_ds["type"].iloc[seg_ds["node_start"]].values
        seg_ds["node_end_type"] = node_ds["type"].iloc[seg_ds["node_end"]].values

        self.nodes = node_ds
        self.segments = seg_ds
        self.disperse_source = ds

    def segment_positions(self) -> FloatArrayType:
        """Compute the segment position.

        Returns
        -------
        pos : np.ndarray, shape (nsegment, 2, ndim)
            The start/end of all segments.
        """
        nseg = self.disperse_source.header.nseg
        ndim = self.disperse_source.header.ndim
        pos = np.asarray(self.disperse_source.data.seg_pos).reshape(nseg, 2, ndim)
        return pos



import numpy as np
import pytest

from ..projections import cic_projection, nearest_projection


@pytest.fixture
def seed():
    return 16091992


def test_projection_conservation(seed):
    np.random.seed(seed)
    # Check that the CIC projection conserves the volume
    X = np.random.rand(1000, 3)

    for proj in (cic_projection, nearest_projection):
        dens = proj(X, 16)
        np.testing.assert_allclose(dens.sum(), len(X))


def test_cic(seed):
    np.random.seed(seed)
    # Corner case
    X = np.array([[0, 0, 0]])
    dens = cic_projection(X, 4)
    dens_real = np.zeros((4, 4, 4))
    for i in (-1, 0):
        for j in (-1, 0):
            for k in (-1, 0):
                dens_real[i, j, k] = 1 / 8

    np.testing.assert_allclose(dens, dens_real)

    # More complicated case -- checked manually
    X = np.random.rand(10, 3)
    dens = cic_projection(X, 4)
    true_ans = [
        dens.ptp(),
        dens.std(),
        dens.min(),
        dens.max(),
        dens[0, 3, 1],
        dens[1, 2, 0],
    ]
    np.testing.assert_allclose(
        true_ans,
        [
            0.800747538955831,
            0.18253974078654206,
            0.0,
            0.800747538955831,
            0.0,
            0.3539705205842787,
        ],
    )



"""Excception classes."""


class AstroToolsetNotSpatialError(Exception):
    def __init__(self, array):
        self.array = array

    def __str__(self):
        msg = (
            "Expected a spatial array, but the last dimension has a shape"
            " of %s instead of 3" % self.array.shape[-1]
        )
        return msg



"""Perform operations on n-dimensional fields."""

import numpy as np

from .funcs import j1_over_x


def gaussian_filter(i, R):
    """Smooth the input on R pixels

    Parameters
    ----------
    i : np.ndarray
        The array to smooth
    R : float
        The sigma parameter of the Gaussian filter

    Returns
    -------
    i_smoothed : np.ndarray
        The input array, smoothed by a Gaussian filter
    """
    ik = np.fft.rfftn(i)
    d = 1 / (2 * np.pi)
    kgrid = np.stack(
        np.meshgrid(
            *[np.fft.fftfreq(_, d=d) for _ in i.shape[:-1]],
            np.fft.rfftfreq(i.shape[-1], d=d),
            indexing="ij",
        ),
        axis=0,
    )
    k2 = np.sum(kgrid ** 2, axis=0)

    return np.fft.irfftn(ik * np.exp(-k2 * (R ** 2 / 2)))


def top_hat_filter(i, R):
    """Smooth the input on R pixels using a Top-Hat filter in Fourier space

    Parameters
    ----------
    i : np.ndarray
        The array to smooth
    R : float
        The radius of the top hat filter

    Returns
    -------
    i_smoothed : np.ndarray
        The input array, smoothed by a Top-Hat
    """
    if i.ndim != 3:
        raise NotImplementedError(
            "Top-Hat filtering has not been implemented for dimensions ≠ 3, got %s."
            % i.ndim
        )
    ik = np.fft.rfftn(i)
    d = 1 / (2 * np.pi)
    kgrid = np.stack(
        np.meshgrid(
            *[np.fft.fftfreq(_, d=d) for _ in i.shape[:-1]],
            np.fft.rfftfreq(i.shape[-1], d=d),
            indexing="ij",
        ),
        axis=0,
    )
    kR = np.linalg.norm(kgrid, axis=0) * R

    W = 3 * j1_over_x(kR)

    return np.fft.irfftn(ik * W)



import numpy as np
import pytest
from scipy.ndimage import gaussian_filter as scipy_gaussian_filter
from scipy.optimize import curve_fit
from scipy.stats import binned_statistic

from ..field_operations import gaussian_filter, top_hat_filter

N1 = 1000
R1 = 50

N3, N3_large = 64, 128
R3 = 20


@pytest.fixture
def x_1D():
    x_1D = np.zeros(N1)
    x_1D[N1 // 2] = 10
    return x_1D


@pytest.fixture
def x_3D():
    x_3D = np.zeros((N3, N3, N3))
    x_3D[N3 // 2, N3 // 2, N3 // 2] = 10
    return x_3D


@pytest.fixture
def x_3D_large():
    x_3D = np.zeros((N3_large, N3_large, N3_large))
    x_3D[N3_large // 2, N3_large // 2, N3_large // 2] = 10
    return x_3D


def test_gaussian_filter_1(x_1D):
    xref = np.arange(N1)
    ynew = gaussian_filter(x_1D, R1)

    def fit_fun(x_1D, C, sigma):
        return C * np.exp(-((x_1D - N1 // 2) ** 2) / 2 / sigma ** 2)

    popt, pcov = curve_fit(fit_fun, xref, ynew, p0=[1, 2])
    Rmeasured = popt[1]
    dR = np.sqrt(pcov[1, 1])

    # Test width of curve is correct
    np.testing.assert_allclose(Rmeasured, R1, rtol=dR, atol=0)


def test_gaussian_filter_compare_to_scipy(x_1D, x_3D):
    for x, R in zip((x_1D, x_3D), (R1, R3)):
        yref = scipy_gaussian_filter(x, R, mode="wrap")
        ynew = gaussian_filter(x, R)

        # Compare result to the one using scipy's gaussian filter
        np.testing.assert_allclose(yref, ynew, atol=1e-4, rtol=0)


def test_gaussian_filter_volume(x_1D, x_3D):
    for x, R in zip((x_1D, x_3D), (R1, R3)):
        ynew = gaussian_filter(x, R)

        # Test volume conservation
        np.testing.assert_allclose(ynew.sum(), x.sum())


def test_gaussian_filter_3d(x_3D):
    yref = scipy_gaussian_filter(x_3D, R3, mode="wrap")
    ynew = gaussian_filter(x_3D, R3)

    # Compare result to the one using scipy's gaussian filter
    np.testing.assert_allclose(yref, ynew, atol=1e-4, rtol=0)


def test_top_hat_filter_exception(x_1D):
    with pytest.raises(NotImplementedError) as excinfo:
        top_hat_filter(x_1D, R1)

    assert "Top-Hat filtering has not been implemented for dimensions ≠ 3" in str(
        excinfo.value
    )


def test_top_hat_filter_radius(x_3D, x_3D_large):
    for x_3D, N in zip((x_3D, x_3D_large), (N3, N3_large)):
        # Compute profile
        Nj = 1j * N
        s = slice(0, N, Nj)
        xyz = np.mgrid[s, s, s] - N / 2
        v = top_hat_filter(x_3D, N / 3)
        r = np.linalg.norm(xyz, axis=0)
        res = binned_statistic(
            r.flatten(), v.flatten(), bins=np.linspace(0, N // 2, 200)
        )

        # Find radius of sphere
        deriv = np.gradient(res.statistic) / np.diff(res.bin_edges)
        Nmin = (res.bin_edges[1:] + res.bin_edges[:-1])[np.nanargmin(deriv)] / 2

        # Can't expect perfect agreement due to grid effects
        np.testing.assert_allclose(Nmin, N / 3, rtol=0.03)


def test_top_hat_filter_volume(x_3D):
    ynew = top_hat_filter(x_3D, R3)
    np.testing.assert_allclose(ynew.sum(), x_3D.sum())



cimport numpy as np

import numpy as np

cimport cython


cdef np.uint64_t[:, :, ::1] state_diagram = np.array([
    1, 2, 3, 2, 4, 5, 3, 5,
    0, 1, 3, 2, 7, 6, 4, 5,
    2, 6, 0, 7, 8, 8, 0, 7,
    0, 7, 1, 6, 3, 4, 2, 5,
    0, 9,10, 9, 1, 1,11,11,
    0, 3, 7, 4, 1, 2, 6, 5,
    6, 0, 6,11, 9, 0, 9, 8,
    2, 3, 1, 0, 5, 4, 6, 7,
    11,11, 0, 7, 5, 9, 0, 7,
    4, 3, 5, 2, 7, 0, 6, 1,
    4, 4, 8, 8, 0, 6,10, 6,
    6, 5, 1, 2, 7, 4, 0, 3,
    5, 7, 5, 3, 1, 1,11,11,
    4, 7, 3, 0, 5, 6, 2, 1,
    6, 1, 6,10, 9, 4, 9,10,
    6, 7, 5, 4, 1, 0, 2, 3,
    10, 3, 1, 1,10, 3, 5, 9,
    2, 5, 3, 4, 1, 6, 0, 7,
    4, 4, 8, 8, 2, 7, 2, 3,
    2, 1, 5, 6, 3, 0, 4, 7,
    7, 2,11, 2, 7, 5, 8, 5,
    4, 5, 7, 6, 3, 2, 0, 1,
    10, 3, 2, 6,10, 3, 4, 4,
    6, 1, 7, 0, 5, 2, 4, 3], dtype=np.uint64).reshape(12, 2, 8).T.copy()

cdef np.int64_t magics[6]
magics[0] = 0x1249249249249249
magics[1] = 0x10c30c30c30c30c3
magics[2] = 0x100f00f00f00f00f
magics[3] = 0x001f0000ff0000ff
magics[4] = 0x001f00000000ffff
magics[5] = 0x1fffff

@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
cdef inline np.int64_t interleave_3bits_64(np.int64_t x) nogil:
    # From https://stackoverflow.com/a/18528775/2601223
    x &= magics[5]
    x = (x | x << 32) & magics[4]
    x = (x | x << 16) & magics[3]
    x = (x | x <<  8) & magics[2]
    x = (x | x <<  4) & magics[1]
    x = (x | x <<  2) & magics[0]
    return x

cdef int128_t magics128[7]
# Set the highest 64 bits first
magics128[0] = 0x924924924924924
magics128[0] <<= 64
magics128[0] |= 0x9249249249249249
magics128[1] = 0x30c30c30c30c30c
magics128[1] <<= 64
magics128[1] |= 0x30c30c30c30c30c3
magics128[2] = 0x300f00f00f00f00
magics128[2] <<= 64
magics128[2] |= 0xf00f00f00f00f00f
magics128[3] = 0x30000ff0000ff00
magics128[3] <<= 64
magics128[3] |= 0xff0000ff0000ff
magics128[4] = 0x3ff00000000
magics128[4] <<= 64
magics128[4] |= 0xffff00000000ffff
magics128[5] = 0x3ff00000000
magics128[5] <<= 64
magics128[5] |= 0xffffffff
magics128[6] = 0x0
magics128[6] <<= 64
magics128[6] |= 0x3ffffffffff

@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
cdef inline int128_t interleave_3bits_128(int128_t x) nogil:
    cdef int128_t y = x
    x &= magics128[6]
    x = (x | x << 64) & magics128[5]
    x = (x | x << 32) & magics128[4]
    x = (x | x << 16) & magics128[3]
    x = (x | x << 8) & magics128[2]
    x = (x | x << 4) & magics128[1]
    x = (x | x << 2) & magics128[0]
    return x


@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
cdef np.uint64_t hilbert3d_single(const np.int64_t X[3], const int bit_length) nogil:
    global state_diagram
    cdef np.uint64_t i_bit_mask, i_bit_mask_out, sdigit, hdigit, cstate, nstate
    cdef int i
    i_bit_mask = ( (interleave_3bits_64(X[0]) << 2)
                 | (interleave_3bits_64(X[1]) << 1)
                 | (interleave_3bits_64(X[2])))
    i_bit_mask_out = 0

    # Build Hilbert ordering using state diagram
    cstate = 0
    for i in range(bit_length-1, -1, -1):
        sdigit = (i_bit_mask >> (3*i)) & 0b111
        nstate = state_diagram[sdigit, 0, cstate]
        hdigit = state_diagram[sdigit, 1, cstate]

        # Set the three bits
        i_bit_mask_out |= hdigit << (3*i)
        cstate = nstate
    return i_bit_mask_out

@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
cdef int128_t hilbert3d_single_quad(const np.int64_t X[3], const int bit_length) nogil:
    global state_diagram
    cdef int128_t i_bit_mask, i_bit_mask_out, sdigit, hdigit, cstate, nstate
    cdef int i
    i_bit_mask = ( (interleave_3bits_128(X[0]) << 2)
                 | (interleave_3bits_128(X[1]) << 1)
                 | (interleave_3bits_128(X[2])))
    i_bit_mask_out = 0

    # Build Hilbert ordering using state diagram
    cstate = 0
    for i in range(bit_length-1, -1, -1):
        sdigit = (i_bit_mask >> (3*i)) & 0b111
        nstate = state_diagram[sdigit, 0, cstate]
        hdigit = state_diagram[sdigit, 1, cstate]

        # Set the three bits
        i_bit_mask_out |= hdigit << (3*i)
        cstate = nstate

    return i_bit_mask_out

@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
cdef np.uint64_t[:] hilbert3d_int(np.int64_t[:, ::1] X, int bit_length):
    global state_diagram
    cdef int npoint, ip
    cdef np.uint64_t[::1] order

    npoint = len(X)
    order = np.zeros(npoint, dtype=np.uint64)

    for ip in range(npoint):
        order[ip] = hilbert3d_single(&X[ip, 0], bit_length)

    return order

@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
def hilbert3d_quad(np.int64_t[:, ::1] X, int bit_length):
    global state_diagram
    cdef int npoint, ip
    cdef np.ndarray[np.float64_t, ndim=1] order

    npoint = len(X)
    order = np.zeros(npoint, dtype=np.float64)

    # This code is rather unfortunate, as the support for float128 in numpy
    # is a bit chaotic
    for ip in range(npoint):
        order[ip] = <np.float64_t> hilbert3d_single_quad(&X[ip, 0], bit_length)

    return order

def hilbert3d(np.ndarray[np.int64_t, ndim=2] X, int bit_length):
    '''Compute the order using Hilbert indexing.

    Arguments
    ---------
    * X: (N, ndim) int ndarray
      The positions
    * bit_length: integer
      The bit_length for the indexing.
    '''

    if bit_length <= 21:
        return np.asarray(hilbert3d_int(X, bit_length), dtype=np.float64)
    elif bit_length <= 42:
        return hilbert3d_quad(X, bit_length)



"""Handle periodicity."""
import numpy as np

from .decorators import spatial
from .types import FloatArrayType


@spatial
def wrap_coordinates(x: FloatArrayType, w: float = 1) -> FloatArrayType:
    """Wrap the position, taking into account periodicity.

    Parameters:
    -----------
    x : 3D array
        The position
    w : float, optional
        The width of the domain (default : 1)

    Returns
    -------
    x_wrapped : 3D array
        The positions, wrapped with periodic boundaries
    """
    x_wrapped = np.mod(x, w)
    return x_wrapped


@spatial
def distance_to_center(
    x: FloatArrayType, w: float = 1, return_center: bool = False
) -> FloatArrayType:
    """Compute the distance to the barycenter, taking into account periodicity.

    Parameters:
    -----------
    x : 3D array
        The position
    w : float, optional
        The width of the domain (default: 1)
    return_center : bool, optional
        Also return the position of the center (default: False)

    Returns
    -------
    dx : 3D array
        The distance to the center, wrapped with periodic boundaries
    xcenter : float array
        The center position, in the same units as x.
        Only returned if return_center is True.
    """
    if np.any(x > w) or np.any(x < 0):
        x = wrap_coordinates(x, w)
    # Compute distances to first element
    x0 = x[..., 0, :]
    dx = x[..., :, :] - x0[..., None, :]
    dx[dx > (+w / 2)] -= w
    dx[dx < (-w / 2)] += w

    # Now rewrite w.r.t. center
    xcenter = dx.mean(axis=-2) + x0
    dx = x - xcenter[..., None, :]
    dx[dx > (+w / 2)] -= w
    dx[dx < (-w / 2)] += w

    if return_center:
        return dx, xcenter
    else:
        return dx



import numpy as np
import pytest
from yt.frontends.ramses.hilbert import hilbert3d as hilbert3d_yt

from astrophysics_toolset.ramses.hilbert import hilbert3d


@pytest.fixture
def seed():
    np.random.seed(16091992)


@pytest.mark.parametrize("bit_length", range(5, 40))
def test_hilbert(seed, bit_length):
    N = 10

    ipos = np.random.randint(2 ** bit_length, size=(N, 3))

    ref = hilbert3d_yt(ipos, bit_length)
    new = hilbert3d(ipos, bit_length).astype(np.float64)

    assert np.any(ref > 0)
    np.testing.assert_allclose(ref, new)



import numpy as np
import yt

dsold = yt.load(
    "/home/ccc/Documents/Postdoc/genetIC-angular-momentum-constrain/simulations/data/DM_256_planck/resim/00140/halo_189/relative_change_lx_0.8/output_00001/info_00001.txt"
)
dsnew = yt.load("output_00080/info_00001.txt")

orderold = np.argsort(dsold.r["io", "particle_identity"])
ordernew = np.argsort(dsnew.r["io", "particle_identity"])

for field in dsnew.field_list:
    print("Testing %s,%s" % field)
    new = dsnew.r[field][ordernew]
    old = dsold.r[field][orderold]

    np.testing.assert_allclose(new, old)



import re
from typing import Iterable, Optional, Sequence

from ..utilities.decorators import read_files
from ..utilities.logging import logger

STRUCT_NAME_RE = re.compile(r"^struct (\w+) \{$")
ARRAY_RE = re.compile(r"^\s*(\w+) (\w+)\((\d+)\);$")
VAR_RE = re.compile(r"^\s*(\w+) (\w+);$")


class PDBReader:
    _known_types = {"int": int, "float": float, "pointer": "yorick_pointer"}

    @read_files(1)
    def __init__(self, fname):
        try:
            import pyorick
        except ModuleNotFoundError as e:
            logger.error(
                "Reading PDB files requires the pyorick package. "
                "Install it via `pip install pyorick` along with Yorick."
            )
            raise e

        if not self.check(fname):
            raise Exception(
                "Unrecognized file format for file %s, "
                "could not read as PDB file." % fname
            )

        self._known_types = self._known_types.copy()
        self.yo = pyorick.Yorick()
        self.yo(f'f=openb("{fname}")')

        self._variables = self._get_vars()
        self._structure = {}
        for v in self._variables:
            self._structure[v] = self._parse_struct(v)

        self._data = {}

    def check(self, fname):
        # Test that the file starts with the magic string
        with open(fname, "br") as f:
            return f.readline().decode() == "!<<PDB:II>>!\n"

    def _get_vars(self):
        self.yo("ptrs=get_vars(f)")

        length = int(self.yo.e("numberof(ptrs)")) - 1
        variables = []
        for i in range(length):
            variables.append(*self.yo.e(f"*ptrs({i+1})"))
        return variables

    def _parse_struct(self, var_name):
        v = f"f.{var_name}"
        lines = self.yo.e(f"print(structof({v}))")

        # Now parse the structure
        remaining = lines[1:-1]  # last line is just }

        variables = {}
        for line in remaining:
            tmp = ARRAY_RE.match(line)
            if tmp:
                type_name, name, length = tmp.groups()
                length = int(length)
            else:
                type_name, name = VAR_RE.match(line).groups()
                length = 0

            if type_name not in self._known_types:
                new_type = self._parse_struct(f"{var_name}.{name}")
                self._known_types[type_name] = new_type
                type_name = new_type
            else:
                type_name = self._known_types[type_name]

            if length == 0:
                spec = type_name
            else:
                spec = (type_name, length)
            variables[name] = spec
        return variables

    @property
    def structure(self):
        return self._structure

    def get(self, path):
        if path in self._data:
            return self._data[path]

        # Check the patch does exist
        keys = path.split("/")
        node = self.structure
        for k in keys:
            node = node.get(k, None)
            if node is None:
                raise KeyError("Provided path does not exist on file: %s" % path)
        cmd = "f.%s" % path.replace("/", ".")

        value = self.yo(f"={cmd}")
        self._data[path] = value
        return value

    def __getitem__(self, key):
        return self.get(key)

    def walk_structure(self, node: dict = None, prefix=None) -> Iterable[list[str]]:
        if node is None:
            node = self.structure
        if prefix is None:
            prefix = []
        for k, v in node.items():
            if isinstance(v, dict):
                yield from self.walk_structure(v, prefix + [k])
                continue

            if isinstance(v, tuple):
                typ, _ = v
            else:
                typ = v
            if typ in (int, float):
                yield prefix + [k]

    def to_hdf(self, filename: str, verbose=False):
        import h5py

        with h5py.File(filename, mode="w") as f:
            for key in ("/".join(_) for _ in self.walk_structure()):
                data = self.get(key)
                logger.info("Reading %s shape=%s", key, data.shape)

                f.create_dataset(key, data=data)


def main(argv: Optional[Sequence[str]] = None) -> int:
    import argparse

    parser = argparse.ArgumentParser("Convert yorick db to hdf5")
    parser.add_argument("fname", help="Yorick file to read")
    parser.add_argument("hdfname", help="HDF5 file to write")
    parser.add_argument("-v", "--verbose", action="store_true")

    args = parser.parse_args(argv)
    if args.verbose:
        logger.setLevel("DEBUG")
    p = PDBReader(args.fname)
    p.to_hdf(args.hdfname)

    return 0


if __name__ == "__main__":
    import sys

    sys.exit(main(sys.argv[1:]))  # pragma: no cover



import logging
import sys

# This next bit is grabbed from:
# http://stackoverflow.com/questions/384076/how-can-i-make-the-python-logging-output-to-be-colored


def add_coloring_to_emit_ansi(fn):
    # add methods we need to the class
    def new(*args):
        levelno = args[0].levelno
        if levelno >= 50:
            color = "\x1b[31m"  # red
        elif levelno >= 40:
            color = "\x1b[31m"  # red
        elif levelno >= 30:
            color = "\x1b[33m"  # yellow
        elif levelno >= 20:
            color = "\x1b[32m"  # green
        elif levelno >= 10:
            color = "\x1b[35m"  # pink
        else:
            color = "\x1b[0m"  # normal
        ln = color + args[0].levelname + "\x1b[0m"
        args[0].levelname = ln
        return fn(*args)

    return new


level = 40
stream = sys.stderr
ufstring = "%(name)-3s: [%(levelname)-9s] %(asctime)s %(message)s"
cfstring = "%(name)-3s: [%(levelname)-18s] %(asctime)s %(message)s"
logger = logging.getLogger("astrophyics_toolset")


def disable_stream_logging():
    if len(logger.handlers) > 0:
        logger.removeHandler(logger.handlers[0])
    h = logging.NullHandler()
    logger.addHandler(h)


def colorize_logging():
    f = logging.Formatter(cfstring)
    logger.handlers[0].setFormatter(f)
    sh.emit = add_coloring_to_emit_ansi(sh.emit)


def uncolorize_logging():
    try:
        f = logging.Formatter(ufstring)
        logger.handlers[0].setFormatter(f)
        sh.emit = original_emitter
    except NameError:
        # sh and original_emitter are not defined because
        # suppressStreamLogging is True, so we continue since there is nothing
        # to uncolorize
        pass


sh = logging.StreamHandler(stream=stream)
# create formatter and add it to the handlers
formatter = logging.Formatter(ufstring)
sh.setFormatter(formatter)
# add the handler to the logger
logger.addHandler(sh)
logger.setLevel(level)
logger.propagate = False

original_emitter = sh.emit

colorize_logging()

logger.debug("Set log level to %s", level)



"""Support Grafic I/O."""

import numpy as np
from scipy.io import FortranFile as FF

from ..utilities.decorators import read_files
from ..utilities.types import PathType
from .common import IOResult


@read_files(1)
def read(fname: PathType) -> IOResult:
    """Read a grafic file.

    Parameters
    ----------
    fname : str, filename

    """
    try:
        aexp = float(fname.split("-")[-1].replace(".dat", ""))
    except ValueError:
        aexp = 1

    with FF(fname, "r") as f:
        N1, N2, N3, qq, Xc1, Xc2, Xc3, Omega_m, Omega_c, Omega_l, H = f.read_record(
            *["i"] * 3, *["f"] * 8
        )
        Lbox = qq * Xc1 * H / 100
        Lbox /= 1000  # Gpc
        Lbox = float(Lbox)

        N1, N2, N3 = (int(_) for _ in (N1, N2, N3))
        Xc = np.stack((Xc1, Xc2, Xc3)).flatten()

        dens = np.zeros((N1, N2, N3), dtype="f")
        for i in range(N1):
            dens[i] = f.read_reals("f").reshape(N2, N3)

    return IOResult(
        data=dict(rho=dens),
        metadata=dict(
            Lbox=Lbox,
            Omega_m=Omega_m,
            Omega_c=Omega_c,
            Omega_l=Omega_l,
            Xc=Xc,
            aexp=aexp,
            N=N1,
        ),
    )



#!/usr/bin/env python

"""The setup script."""

import numpy as np
from Cython.Build import cythonize
from setuptools import Extension, setup

extensions = [
    Extension(
        "**/*",
        ["**/*.pyx"],
        include_dirs=[np.get_include()],
        define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
    )
]

setup(
    ext_modules=cythonize(
        extensions, language_level=3, annotate=True, gdb_debug=True, nthreads=4
    ),
)



import yt
from yt.data_objects.selection_objects.data_selection_objects import (
    YTSelectionContainer,
)
from yt.data_objects.static_output import Dataset

from ..utilities.logging import logger


def create_sph_fields(
    data_source: YTSelectionContainer, ptype: str = "all", *args, **kwargs
) -> Dataset:
    """
    Return a dataset with sph interpolation

    Parameters:
    -----------
    data_source : YTSelectionContainer
        The data source from which particles will be extracted
    ptype : str
        The type of particle to extract
    args, kwargs : extra arguments will be passed to `add_sph_fields`

    Returns
    -------
    ds : Dataset
        A dataset with sph interpolation.
    """
    ds = data_source.ds

    if hasattr(ds, "add_sph_fields"):
        ds.add_sph_fields(*args, **kwargs)
        return ds

    logger.debug("Loading data from data source")
    data = {
        (ptype, fname): data_source[ftype, fname]
        for (ftype, fname) in ds.field_list
        if (ftype == ptype) and (data_source[ftype, fname].ndim == 1)
    }

    # Make sure position and masses are copied
    for k in "xyz":
        data[ptype, f"particle_position_{k}"] = data_source[
            ptype, f"particle_position_{k}"
        ]
    data[ptype, "particle_mass"] = data_source[ptype, "particle_mass"]
    data[ptype, "particle_identity"] = data_source[ptype, "particle_identity"]
    data[ptype, "particle_family"] = data_source[ptype, "particle_family"]

    periodicity = (True, True, True)
    # No support for aperiodicity for now
    # L, R = data_source.get_bbox()
    # periodicity = (
    #     np.asarray(ds.periodicity)
    #     & (ds.domain_left_edge == L)
    #     & (ds.domain_right_edge == R)
    # )

    logger.debug("Create particle dataset")
    sph_ds = yt.load_particles(data, data_source=data_source, periodicity=periodicity)
    sph_ds.current_redshift = ds.current_redshift  # Copy current redshift
    sph_ds.add_sph_fields(*args, **kwargs, sph_ptype=ptype)

    return sph_ds
